# coding=utf-8

import os
from tqdm import tqdm
os.environ["CUDA_VISIBLE_DEVICES"] = "0"
import numpy as np
import tensorflow as tf
# tf.enable_eager_execution()

x = tf.Variable(tf.random.normal([1000, 20]))

index = np.random.randint(0, 1000, 20000)

import time
start_time = time.time()

@tf.function
def test():
    y = tf.gather(x, index)
    return y

for _ in tqdm(range(10000)):
    z = test()

test()

end_time = time.time()
print("elapsed: {}".format(end_time - start_time))

import tf_geometric as tfg



# coding=utf-8
import tensorflow as tf
from tf_geometric.nn.conv.gin import gin


class GIN(tf.keras.Model):
    """
    Graph Isomorphism Network  Layer
    """

    def __init__(self, mlp_model, eps=0, train_eps=False, *args, **kwargs):
        """
        :param mlp_model: A neural network (multi-layer perceptrons).
        :param eps: float, optional, (default: :obj:`0.`).
        :param train_eps: Boolean, Whether the eps is trained.
        :param activation: Activation function to use.
        """
        super().__init__(*args, **kwargs)
        self.mlp_model = mlp_model

        self.eps = eps
        if train_eps:
            self.eps = self.add_weight("eps", shape=[], initializer="zeros")


    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GIN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return gin(x, edge_index, edge_weight, self.mlp_model, self.eps, training=training)



# coding=utf-8


import tensorflow as tf
from tf_geometric.nn.pool.asap import asap


class ASAP(tf.keras.Model):
    """
    OOP API for ASAP: Adaptive Structure Aware Pooling for Learning Hierarchical Graph Representation
    """

    def __init__(self, K=None, ratio=None, drop_rate=0.0,
                 attention_units=None,
                 le_conv_activation=tf.nn.sigmoid,
                 le_conv_use_bias=True,
                 kernel_regularizer=None, bias_regularizer=None,
                 *args, **kwargs):
        """
        ASAPool
        :param attention_units: Positive integer, dimensionality for attention.
        """
        super().__init__(*args, **kwargs)

        self.attention_units = attention_units

        self.attention_gcn_kernel = None
        self.attention_gcn_bias = None
        self.attention_query_kernel = None
        self.attention_query_bias = None
        self.attention_score_kernel = None
        self.attention_score_bias = None

        self.le_conv_self_kernel = None
        self.le_conv_self_bias = None
        self.le_conv_aggr_self_kernel = None
        self.le_conv_aggr_self_bias = None
        self.le_conv_aggr_neighbor_kernel = None

        self.K = K
        self.ratio = ratio
        self.drop_rate = drop_rate
        self.le_conv_activation = le_conv_activation

        self.le_conv_use_bias = le_conv_use_bias

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

    def build(self, input_shapes):
        x_shape = input_shapes[0]
        num_features = x_shape[-1]

        if self.attention_units is None:
            self.attention_units = num_features



        self.attention_gcn_kernel = self.add_weight("attention_gcn_kernel", shape=[num_features, self.attention_units],
                                                    initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        self.attention_gcn_bias = self.add_weight("attention_gcn_bias", shape=[self.attention_units],
                                        initializer="zeros", regularizer=self.bias_regularizer)
        self.attention_query_kernel = self.add_weight("attention_query_kernel", shape=[self.attention_units, self.attention_units],
                                                    initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        self.attention_query_bias = self.add_weight("attention_query_bias", shape=[self.attention_units],
                                        initializer="zeros", regularizer=self.bias_regularizer)
        self.attention_score_kernel = self.add_weight("attention_score_kernel", shape=[self.attention_units * 2, 1],
                                                    initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        self.attention_score_bias = self.add_weight("attention_score_bias", shape=[1],
                                        initializer="zeros", regularizer=self.bias_regularizer)

        self.le_conv_self_kernel = self.add_weight("le_conv_self_kernel", shape=[self.attention_units, 1],
                                                   initializer="glorot_uniform", regularizer=self.kernel_regularizer)

        if self.le_conv_use_bias:
            self.le_conv_self_bias = self.add_weight("le_conv_self_bias", shape=[1],
                                                     initializer="zeros", regularizer=self.bias_regularizer)

        self.le_conv_aggr_self_kernel = self.add_weight("le_conv_aggr_self_kernel", shape=[self.attention_units, 1],
                                                        initializer="glorot_uniform", regularizer=self.kernel_regularizer)

        if self.le_conv_use_bias:
            self.le_conv_aggr_self_bias = self.add_weight("le_conv_aggr_self_bias", shape=[1],
                                                      initializer="zeros", regularizer=self.bias_regularizer)

        self.le_conv_aggr_neighbor_kernel = self.add_weight("le_conv_aggr_neighbor_kernel", shape=[self.attention_units, 1],
                                                            initializer="glorot_uniform", regularizer=self.kernel_regularizer)


    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight, node_graph_index]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """
        x, edge_index, edge_weight, node_graph_index = inputs

        return asap(x, edge_index, edge_weight, node_graph_index,
                        self.attention_gcn_kernel, self.attention_gcn_bias,
                        self.attention_query_kernel, self.attention_query_bias,
                        self.attention_score_kernel, self.attention_score_bias,
                        self.le_conv_self_kernel, self.le_conv_self_bias,
                        self.le_conv_aggr_self_kernel, self.le_conv_aggr_self_bias,
                        self.le_conv_aggr_neighbor_kernel, None,
                        K=self.K, ratio=self.ratio, le_conv_activation=self.le_conv_activation,
                        drop_rate=self.drop_rate, training=training, cache=cache)



# coding=utf-8
import tensorflow as tf
from tf_geometric.nn.kernel.map_reduce import aggregate_neighbors, sum_reducer, identity_updater
from tf_geometric.utils.graph_utils import add_self_loop_edge


CACHE_KEY_GCN_NORMED_EDGE_TEMPLATE = "gcn_normed_edge_{}_{}"


def compute_cache_key(renorm, improved):
    """
    Compute the cached key based on GCN normalization configurations: renorm and improved

    :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
    :param improved: Whether use improved GCN or not.
    :return: The corresponding cached key for the given GCN normalization configuration.
    """
    return CACHE_KEY_GCN_NORMED_EDGE_TEMPLATE.format(renorm, improved)


def gcn_norm_edge(edge_index, num_nodes, edge_weight=None, renorm=True, improved=False, cache: dict=None):
    """
    Compute normed edge (updated edge_index and normalized edge_weight) for GCN normalization.

    :param edge_index: Tensor, shape: [2, num_edges], edge information.
    :param num_nodes: Number of nodes.
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
    :param improved: Whether use improved GCN or not.
    :param cache: A dict for caching the updated edge_index and normalized edge_weight.
    :return: Normed edge (updated edge_index and normalized edge_weight).
    """

    if cache is not None:
        cache_key = compute_cache_key(renorm, improved)
        cached_data = cache.get(cache_key, None)
        if cached_data is not None:
            return cached_data

    if edge_weight is None:
        edge_weight = tf.ones([tf.shape(edge_index)[1]], dtype=tf.float32)

    fill_weight = 2.0 if improved else 1.0

    if renorm:
        edge_index, edge_weight = add_self_loop_edge(edge_index, num_nodes, edge_weight=edge_weight, fill_weight=fill_weight)

    row, col = edge_index[0], edge_index[1]
    deg = tf.math.unsorted_segment_sum(edge_weight, row, num_segments=num_nodes)
    deg_inv_sqrt = tf.pow(deg, -0.5)
    deg_inv_sqrt = tf.where(
        tf.math.logical_or(tf.math.is_inf(deg_inv_sqrt), tf.math.is_nan(deg_inv_sqrt)),
        tf.zeros_like(deg_inv_sqrt),
        deg_inv_sqrt
    )

    normed_edge_weight = tf.gather(deg_inv_sqrt, row) * edge_weight * tf.gather(deg_inv_sqrt, col)

    if not renorm:
        edge_index, normed_edge_weight = add_self_loop_edge(edge_index, num_nodes, edge_weight=normed_edge_weight,
                                                            fill_weight=fill_weight)

    if cache is not None:
        cache[cache_key] = edge_index, normed_edge_weight

    return edge_index, normed_edge_weight


def gcn_cache_normed_edge(graph, renorm=True, improved=False, override=False):
    """
    Manually compute the normed edge based on the given GCN normalization configuration (renorm and improved) and put it in graph.cache.
    If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

    :param graph: tfg.Graph, the input graph.
    :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
    :param improved: Whether use improved GCN or not.
    :param override: Whether to override existing cached normed edge.
    :return: None
    """
    if override:
        cache_key = compute_cache_key(renorm, improved)
        graph.cache[cache_key] = None
    gcn_norm_edge(graph.edge_index, graph.num_nodes, graph.edge_weight, renorm, improved, graph.cache)


def gcn_mapper(repeated_x, neighbor_x, edge_weight=None):
    return neighbor_x * tf.expand_dims(edge_weight, 1)


def gcn(x, edge_index, edge_weight, kernel, bias=None, activation=None,
        renorm=True, improved=False, cache=None):
    """
    Functional API for Graph Convolutional Networks.

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param kernel: Tensor, shape: [num_features, num_output_features], weight
    :param bias: Tensor, shape: [num_output_features], bias
    :param activation: Activation function to use.
    :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
    :param improved: Whether use improved GCN or not.
    :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        To use @tf_utils.function with gcn, you should cache the noremd edge information before the first call of the gcn.

        - (1) If you're using OOP APIs tfg.layers.GCN:

              gcn_layer.cache_normed_edge(graph)

        - (2) If you're using functional API tfg.nn.gcn:

              from tf_geometric.nn.conv.gcn import gcn_cache_normed_edge
              gcn_cache_normed_edge(graph)

    :return: Updated node features (x), shape: [num_nodes, num_output_features]
    """

    num_nodes = tf.shape(x)[0]
    updated_edge_index, normed_edge_weight = gcn_norm_edge(edge_index, num_nodes, edge_weight, renorm, improved, cache)

    h = x @ kernel

    h = aggregate_neighbors(
        h, updated_edge_index, normed_edge_weight,
        gcn_mapper,
        sum_reducer,
        identity_updater,
        num_nodes=num_nodes
    )

    if bias is not None:
        h += bias

    if activation is not None:
        h = activation(h)

    return h






# coding=utf-8
import tensorflow as tf

if tf.__version__[0] == "1":
    tf.enable_eager_execution()

import tf_geometric.nn as nn
import tf_geometric.utils as utils
import tf_geometric.data as data
import tf_geometric.datasets as datasets
import tf_geometric.layers as layers

from tf_geometric.data.graph import Graph, BatchGraph






# coding=utf-8

import tensorflow as tf
from tf_geometric.nn.conv.graph_sage import mean_graph_sage, gcn_graph_sage, mean_pool_graph_sage, \
    max_pool_graph_sage, lstm_graph_sage, sum_graph_sage


class MeanGraphSage(tf.keras.Model):
    """
    GraphSAGE: `"Inductive Representation Learning on Large Graphs" <https://arxiv.org/abs/1706.02216>`_ paper
    """

    def __init__(self, units, activation=tf.nn.relu, use_bias=True, concat=True,
                 normalize=False,
                 kernel_regularizer=None, bias_regularizer=None,
                 *args, **kwargs):
        """

        :param units:
        :param activation:
        :param use_bias:
        :param concat:
        :param normalize:
        :param kernel_regularizer: Regularizer function applied to the `kernel` weights matrix.
        :param bias_regularizer: Regularizer function applied to the bias vector.
        :param args:
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.concat = concat
        self.normalize = normalize

        if concat and (units % 2 != 0):
            raise Exception("units must be a event number if concat is True")

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

        self.self_kernel = None
        self.neighbor_kernel = None
        self.bias = None

    def build(self, input_shape):
        x_shape = input_shape[0]
        num_features = x_shape[-1]

        if self.concat:
            kernel_units = self.units // 2
        else:
            kernel_units = self.units

        self.self_kernel = self.add_weight("self_kernel", shape=[num_features, kernel_units],
                                           initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        self.neighbor_kernel = self.add_weight("neighbor_kernel", shape=[num_features, kernel_units],
                                               initializer="glorot_uniform", regularizer=self.kernel_regularizer)

        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units], initializer="zeros", regularizer=self.bias_regularizer)

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """
        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return mean_graph_sage(x, edge_index, edge_weight,
                               self.self_kernel,
                               self.neighbor_kernel,
                               bias=self.bias,
                               activation=self.activation, concat=self.concat,
                               normalize=self.normalize)

class SumGraphSage(tf.keras.Model):
    """
    GraphSAGE: `"Inductive Representation Learning on Large Graphs" <https://arxiv.org/abs/1706.02216>`_ paper
    """

    def __init__(self, units, activation=tf.nn.relu, use_bias=True, concat=True,
                 normalize=False,
                 kernel_regularizer=None, bias_regularizer=None,
                 *args, **kwargs):
        """

        :param units:
        :param activation:
        :param use_bias:
        :param concat:
        :param normalize:
        :param kernel_regularizer: Regularizer function applied to the `kernel` weights matrix.
        :param bias_regularizer: Regularizer function applied to the bias vector.
        :param args:
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.concat = concat
        self.normalize = normalize

        if concat and (units % 2 != 0):
            raise Exception("units must be a event number if concat is True")

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

        self.self_kernel = None
        self.neighbor_kernel = None
        self.bias = None

    def build(self, input_shape):
        x_shape = input_shape[0]
        num_features = x_shape[-1]

        if self.concat:
            kernel_units = self.units // 2
        else:
            kernel_units = self.units

        self.self_kernel = self.add_weight("self_kernel", shape=[num_features, kernel_units],
                                           initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        self.neighbor_kernel = self.add_weight("neighbor_kernel", shape=[num_features, kernel_units],
                                               initializer="glorot_uniform", regularizer=self.kernel_regularizer)

        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units], initializer="zeros", regularizer=self.bias_regularizer)

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """
        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return sum_graph_sage(x, edge_index, edge_weight,
                               self.self_kernel,
                               self.neighbor_kernel,
                               bias=self.bias,
                               activation=self.activation, concat=self.concat,
                               normalize=self.normalize)


class GCNGraphSage(tf.keras.Model):
    def __init__(self, units, activation=tf.nn.relu, use_bias=True,
                 normalize=False,
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.normalize = normalize

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

        self.kernel = None
        self.bias = None

    def build(self, input_shape):
        x_shape = input_shape[0]
        num_features = x_shape[-1]

        self.kernel = self.add_weight("kernel", shape=[num_features, self.units],
                                      initializer="glorot_uniform", regularizer=self.kernel_regularizer)

        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units],
                                        initializer="zeros", regularizer=self.bias_regularizer)

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return gcn_graph_sage(x, edge_index, edge_weight, self.kernel, self.bias, self.activation, self.normalize,
                              cache=cache)


class MeanPoolGraphSage(tf.keras.Model):
    def __init__(self, units, activation=tf.nn.relu, use_bias=True, concat=True,
                 normalize=False,
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.concat = concat

        if concat and (units % 2 != 0):
            raise Exception("units must be a event number if concat is True")

        self.normalize = normalize

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

        self.self_kernel = None
        self.neighbor_mlp_kernel = None
        self.neighbor_mlp_bias = None
        self.neighbor_kernel = None


        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units],
                                        initializer="zeros", regularizer=self.bias_regularizer)

    def build(self, input_shape):
        x_shape = input_shape[0]
        num_features = x_shape[-1]

        if self.concat:
            kernel_units = self.units // 2
        else:
            kernel_units = self.units

        self.self_kernel = self.add_weight("self_kernel", shape=[num_features, kernel_units],
                                           initializer="glorot_uniform", regularizer=self.kernel_regularizer)

        self.neighbor_mlp_kernel = self.add_weight("neighbor_mlp_kernel", shape=[num_features, kernel_units * 4],
                                                   initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        if self.use_bias:
            self.neighbor_mlp_bias = self.add_weight("neighbor_mlp_bias", shape=[kernel_units * 4],
                                                     initializer="zeros", regularizer=self.bias_regularizer)

        self.neighbor_kernel = self.add_weight("neighbor_kernel", shape=[kernel_units * 4, kernel_units],
                                               initializer="glorot_uniform", regularizer=self.kernel_regularizer)

        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units],
                                        initializer="zeros", regularizer=self.bias_regularizer)

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return mean_pool_graph_sage(x, edge_index, edge_weight,
                                    self.self_kernel,
                                    self.neighbor_mlp_kernel,
                                    self.neighbor_kernel,
                                    neighbor_mlp_bias=self.neighbor_mlp_bias,
                                    bias=self.bias, activation=self.activation,
                                    concat=self.concat, normalize=self.normalize)


class MaxPoolGraphSage(tf.keras.Model):
    def __init__(self, units, activation=tf.nn.relu, use_bias=True,
                 concat=True, normalize=False,
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.concat = concat

        if concat and (units % 2 != 0):
            raise Exception("units must be a event number if concat is True")

        self.normalize = normalize

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

        self.self_kernel = None
        self.neighbor_mlp_kernel = None
        self.neighbor_mlp_bias = None
        self.neighbor_kernel = None
        self.bias = None

    def build(self, input_shape):
        x_shape = input_shape[0]
        num_features = x_shape[-1]

        if self.concat:
            kernel_units = self.units // 2
        else:
            kernel_units = self.units

        self.self_kernel = self.add_weight("self_kernel", shape=[num_features, kernel_units],
                                           initializer="glorot_uniform", regularizer=self.kernel_regularizer)

        self.neighbor_mlp_kernel = self.add_weight("mlp_kernel", shape=[num_features, kernel_units * 4],
                                                   initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        if self.use_bias:
            self.neighbor_mlp_bias = self.add_weight("mlp_bias", shape=[kernel_units * 4],
                                                     initializer="zeros", regularizer=self.bias_regularizer)
        self.neighbor_kernel = self.add_weight("neighs_kernel", shape=[kernel_units * 4, kernel_units],
                                               initializer="glorot_uniform", regularizer=self.kernel_regularizer)

        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units],
                                        initializer="zeros", regularizer=self.bias_regularizer)

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return max_pool_graph_sage(x, edge_index, edge_weight,
                                   self.self_kernel,
                                   self.neighbor_mlp_kernel,
                                   self.neighbor_kernel,
                                   neighbor_mlp_bias=self.neighbor_mlp_bias,
                                   bias=self.bias, activation=self.activation,
                                   concat=self.concat, normalize=self.normalize)


class LSTMGraphSage(tf.keras.Model):

    def __init__(self, units, activation=tf.nn.relu, use_bias=True,
                 concat=True, normalize=False,
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.concat = concat
        self.normalize = normalize

        if concat and (units % 2 != 0):
            raise Exception("units must be a event number if concat is True")

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

        self.lstm = None
        self.self_kernel = None
        self.neighbor_kernel = None
        self.bias = None

    def build(self, input_shape):
        x_shape = input_shape[0]
        num_features = x_shape[-1]

        if self.concat:
            kernel_units = self.units // 2
        else:
            kernel_units = self.units

        self.lstm = tf.keras.layers.LSTM(kernel_units, return_sequences=True,
                                         kernel_regularizer=self.kernel_regularizer,
                                         bias_regularizer=self.bias_regularizer)

        self.self_kernel = self.add_weight("self_kernel", shape=[num_features, kernel_units],
                                           initializer="glorot_uniform", regularizer=self.kernel_regularizer)

        self.neighbor_kernel = self.add_weight("neighbor_kernel", shape=[kernel_units, kernel_units],
                                               initializer="glorot_uniform", regularizer=self.kernel_regularizer)


        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units],
                                        initializer="zeros", regularizer=self.bias_regularizer)

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index] or [x, edge_index, edge_weight].
            Note that the edge_weight will not be used.
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        x, edge_index = inputs[0], inputs[1]

        return lstm_graph_sage(x, edge_index, self.lstm,
                               self.self_kernel,
                               self.neighbor_kernel,
                               bias=self.bias, activation=self.activation,
                               concat=self.concat, normalize=self.normalize)



# coding=utf-8
import tensorflow as tf
from tf_geometric.data.graph import BatchGraph
from tf_geometric.nn.pool.topk_pool import topk_pool


def sag_pool(x, edge_index, edge_weight, node_graph_index,
             score_gnn, K=None, ratio=None,
             score_activation=None, training=None, cache=None):
    """
    Functional API for SAGPool

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param node_graph_index: Tensor/NDArray, shape: [num_nodes], graph index for each node
    :param score_gnn: A GNN model to score nodes for the pooling, [x, edge_index, edge_weight] => node_score.
    :param K: Keep top K targets for each source
    :param ratio: Keep num_targets * ratio targets for each source
    :param score_activation: Activation to use for node_score before multiplying node_features with node_score
    :param training: Python boolean indicating whether the layer should behave in
        training mode (adding dropout) or in inference mode (doing nothing).
    :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
    :return: [pooled_x, pooled_edge_index, pooled_edge_weight, pooled_node_graph_index]
    """

    if cache is None:
        node_score = score_gnn([x, edge_index, edge_weight], training=training)
    else:
        node_score = score_gnn([x, edge_index, edge_weight], training=training, cache=cache)

    topk_node_index = topk_pool(node_graph_index, node_score, K=K, ratio=ratio)

    if score_activation is not None:
        node_score = score_activation(node_score)

    pooled_graph = BatchGraph(
        x=x * node_score,
        edge_index=edge_index,
        node_graph_index=node_graph_index,
        edge_graph_index=None,
        edge_weight=edge_weight
    ).sample_new_graph_by_node_index(topk_node_index)

    return pooled_graph.x, pooled_graph.edge_index, pooled_graph.edge_weight, pooled_graph.node_graph_index



# coding=utf-8

import tensorflow as tf
import warnings


def warn_tf1():
    warnings.warn("Using @tf_utils.function to speed-up functions is only available for TensorFlow 2.x. "
                  "Upgrade your TensorFlow to 2.x and the performance can be improved dramatically with @tf_utils.function")


def tf_func_warn(*args, **kwargs):
    if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
        func = args[0]
        warn_tf1()
        return func
    else:
        def decorate(func):
            warn_tf1()
            return func
        return decorate


# disable tf.function for tf 1
if tf.__version__[0] == "1":
    function = tf_func_warn
else:
    function = tf.function



# coding=utf-8
import os
# multi-gpu ids
os.environ["CUDA_VISIBLE_DEVICES"] = "4"
import tf_geometric as tfg
from tf_geometric.layers import GCN
from tensorflow.keras.regularizers import L1L2
import tensorflow as tf


graph, (train_index, valid_index, test_index) = tfg.datasets.CoraDataset().load_data()
num_classes = graph.y.max() + 1
learning_rate = 1e-2
l2_coef = 5e-4

model = tfg.layers.GAT(units=5)
model.compile(
    optimizer=tf.keras.optimizers.Adam(learning_rate=1e-1),
    loss=tf.keras.losses.categorical_crossentropy
)

print(model([graph.x, graph.edge_index]))


from scipy.sparse.linalg import eigs, eigsh
from tf_geometric.utils.graph_utils import get_laplacian, to_scipy_sparse_matrix, remove_self_loop_edge, add_self_loop_edge


class LaplacianLambdaMax(object):
    r"""Computes the highest eigenvalue of the graph Laplacian given by
    :meth:`torch_geometric.utils.get_laplacian`.

    Args:
        normalization (str, optional): The normalization scheme for the graph
            Laplacian (default: :obj:`None`):

            1. :obj:`None`: No normalization
            :math:`\mathbf{L} = \mathbf{D} - \mathbf{A}`

            2. :obj:`"sym"`: Symmetric normalization
            :math:`\mathbf{L} = \mathbf{I} - \mathbf{D}^{-1/2} \mathbf{A}
            \mathbf{D}^{-1/2}`

            3. :obj:`"rw"`: Random-walk normalization
            :math:`\mathbf{L} = \mathbf{I} - \mathbf{D}^{-1} \mathbf{A}`
        is_undirected (bool, optional): If set to :obj:`True`, this transform
            expects undirected graphs as input, and can hence speed up the
            computation of the largest eigenvalue. (default: :obj:`False`)
    """

    def __init__(self, normalization_type=None, is_undirected=False):
        assert normalization_type in [None, 'sym', 'rw'], 'Invalid normalization'
        self.normalization = normalization_type
        self.is_undirected = is_undirected

    def __call__(self, data):
        edge_index = data.edge_index
        edge_weight = data.edge_weight

        edge_index, edge_weight = remove_self_loop_edge(edge_index, edge_weight)

        edge_index, edge_weight = get_laplacian(data.edge_index, data.x.shape[0], edge_weight,
                                                self.normalization)
        # edge_index, edge_weight = add_self_loop_edge(edge_index, data.x.shape[0], edge_weight, fill_weight=-1.)

        L = to_scipy_sparse_matrix(edge_index, edge_weight, data.x.shape[0])

        eig_fn = eigs
        if self.is_undirected and self.normalization != 'rw':
            eig_fn = eigsh

        lambda_max = eig_fn(L, k=1, which='LM', return_eigenvectors=False)
        data.lambda_max = float(lambda_max)

        return data

    def __repr__(self):
        return '{}(normalization={})'.format(self.__class__.__name__,
                                             self.normalization)



# coding=utf-8

import tensorflow as tf

from tf_geometric.nn.pool.sort_pool import sort_pool


class SortPool(tf.keras.Model):
    """
    OOP API for SortPool "An End-to-End Deep Learning Architecture for Graph Classification"
    """

    def __init__(self, K=None, ratio=None, sort_index=-1,  *args, **kwargs):
        """
        SAGPool

        :param score_gnn: A GNN model to score nodes for the pooling, [x, edge_index, edge_weight] => node_score.
        :param K: Keep top K targets for each source
        :param ratio: Keep num_targets * ratio targets for each source
        :param sort_index: The sort_index_th index of the last axis will used for sort.
        """
        super().__init__(*args, **kwargs)
        self.K = K
        self.ratio = ratio
        self.sort_index = sort_index

    def call(self, inputs, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight, node_graph_index]
        :return: Pooled grpah: [pooled_x, pooled_edge_index, pooled_edge_weight, pooled_node_graph_index]
        """
        x, edge_index, edge_weight, node_graph_index = inputs

        return sort_pool(x, edge_index, edge_weight, node_graph_index,
                         K=self.K, ratio=self.ratio, sort_index=self.sort_index, training=training)



# coding=utf-8
import tensorflow as tf
from tf_geometric.nn.pool.cluster_pool import cluster_pool

from tf_geometric.nn.conv.le_conv import le_conv
from tf_geometric.nn.kernel.map_reduce import aggregate_neighbors, sum_reducer, identity_updater, identity_mapper, \
    max_reducer

from tf_geometric.nn.kernel.segment import segment_softmax

from tf_geometric.nn import max_pool
from tf_geometric.nn.conv.gcn import gcn, gcn_mapper

from tf_geometric.data.graph import BatchGraph
from tf_geometric.nn.pool.topk_pool import topk_pool
from tf_geometric.utils.graph_utils import add_self_loop_edge, remove_self_loop_edge


def asap(x, edge_index, edge_weight, node_graph_index,
         attention_gcn_kernel, attention_gcn_bias,
         attention_query_kernel, attention_query_bias,
         attention_score_kernel, attention_score_bias,
         le_conv_self_kernel, le_conv_self_bias,
         le_conv_aggr_self_kernel, le_conv_aggr_self_bias,
         le_conv_aggr_neighbor_kernel, le_conv_aggr_neighbor_bias,
         K=None, ratio=None,
         le_conv_activation=tf.nn.sigmoid,
         drop_rate=0.0, training=None, cache=None):
    """
    Functional API for ASAP: Adaptive Structure Aware Pooling for Learning Hierarchical Graph Representation

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param node_graph_index: Tensor/NDArray, shape: [num_nodes], graph index for each node
    :param K: Keep top K targets for each source
    :param ratio: Keep num_targets * ratio targets for each source
    :param le_conv_activation: Activation to use for node_score before multiplying node_features with node_score
    :param training: Python boolean indicating whether the layer should behave in
        training mode (adding dropout) or in inference mode (doing nothing).
    :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
    :return: [pooled_x, pooled_edge_index, pooled_edge_weight, pooled_node_graph_index]
    """

    num_nodes = tf.shape(x)[0]
    # num_graphs = tf.reduce_max(node_graph_index) + 1

    edge_index, edge_weight = remove_self_loop_edge(edge_index, edge_weight)
    edge_index_with_self_loop, edge_weight_with_self_loop = add_self_loop_edge(edge_index, num_nodes=num_nodes, edge_weight=edge_weight)

    row_with_self_loop, col_with_self_loop = edge_index_with_self_loop[0], edge_index_with_self_loop[1]

    attention_h = gcn(x, edge_index, edge_weight, attention_gcn_kernel, attention_gcn_bias, cache=cache)

    # max_pool -> query
    attention_query = aggregate_neighbors(
        attention_h, edge_index_with_self_loop, None,
        mapper=identity_mapper,
        reducer=max_reducer,
        updater=identity_updater,
        num_nodes=num_nodes
    )

    attention_query = attention_query @ attention_query_kernel + attention_query_bias

    repeated_attention_query = tf.gather(attention_query, row_with_self_loop)
    repeated_attention_h = tf.gather(attention_h, col_with_self_loop)

    attention_score_h = tf.concat([repeated_attention_query, repeated_attention_h], axis=-1)
    attention_score = attention_score_h @ attention_score_kernel + attention_score_bias
    attention_score = tf.nn.leaky_relu(attention_score, alpha=0.2)

    normed_attention_score = segment_softmax(attention_score, row_with_self_loop, num_nodes)
    if training and drop_rate > 0:
        normed_attention_score = tf.compat.v2.nn.dropout(normed_attention_score, rate=drop_rate)

    # nodes are clusters
    cluster_h = aggregate_neighbors(
        x, edge_index_with_self_loop, tf.reshape(normed_attention_score, [-1]),
        gcn_mapper,
        sum_reducer,
        identity_updater,
        num_nodes=num_nodes
    )

    node_score = le_conv(cluster_h, edge_index, edge_weight,
                         le_conv_self_kernel, le_conv_self_bias,
                         le_conv_aggr_self_kernel, le_conv_aggr_self_bias,
                         le_conv_aggr_neighbor_kernel, le_conv_aggr_neighbor_bias,
                         activation=None)


    topk_node_index = topk_pool(node_graph_index, node_score, K=K, ratio=ratio)
    topk_node_score = tf.gather(node_score, topk_node_index)
    if le_conv_activation is not None:
        topk_node_score = le_conv_activation(topk_node_score)

    pooled_x = tf.gather(cluster_h, topk_node_index) * topk_node_score

    num_clusters = tf.shape(topk_node_index)[0]
    # node->cluster
    cluster_reverse_index = tf.cast(tf.fill([num_nodes], -1), tf.int32)
    cluster_reverse_index = tf.tensor_scatter_nd_update(
        cluster_reverse_index,
        tf.expand_dims(topk_node_index, axis=-1),
        tf.range(num_clusters)
    )

    # row, col = edge_index[0], edge_index[1]
    assign_row = tf.gather(cluster_reverse_index, row_with_self_loop)
    assign_mask = tf.greater_equal(assign_row, 0)

    assign_row = tf.boolean_mask(assign_row, assign_mask)
    assign_col = tf.boolean_mask(col_with_self_loop, assign_mask)
    assign_edge_index = tf.stack([assign_row, assign_col], axis=0)

    assign_edge_weight = tf.boolean_mask(normed_attention_score, assign_mask)
    assign_edge_weight = tf.reshape(assign_edge_weight, [-1])
    assign_edge_weight = tf.stop_gradient(assign_edge_weight)


    # Coarsen in a large BatchGraph.
    _, pooled_edge_index, pooled_edge_weight = cluster_pool(
        None, edge_index_with_self_loop, edge_weight_with_self_loop, assign_edge_index, assign_edge_weight,
        num_clusters, num_nodes=num_nodes)

    pooled_edge_index, pooled_edge_weight = remove_self_loop_edge(pooled_edge_index, pooled_edge_weight)
    pooled_edge_index, pooled_edge_weight = add_self_loop_edge(pooled_edge_index, num_clusters, pooled_edge_weight)

    pooled_node_graph_index = tf.gather(node_graph_index, topk_node_index)

    return pooled_x, pooled_edge_index, pooled_edge_weight, pooled_node_graph_index



# coding=utf-8
import tensorflow as tf

from tf_geometric.utils.graph_utils import convert_dense_adj_to_edge, remove_self_loop_edge, add_self_loop_edge


def cluster_pool(x, edge_index, edge_weight, assign_edge_index, assign_edge_weight, num_clusters, num_nodes=None):
    """
    Coarsen the input Graph based on cluster assignment of nodes and output pooled Graph.

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param assign_edge_index: Tensor, shape: [2, num_nodes], edge between clusters and nodes, where each edge
        denotes a node belongs to a specific cluster.
    :param assign_edge_weight: Tensor or None, shape: [num_nodes], the corresponding weight for assign_edge_index
    :param num_clusters: Number of clusters.
    :param num_nodes: Number of nodes, Optional, used for boosting performance.
    :return: Pooled Graph: [pooled_x, pooled_edge_index, pooled_edge_weight]
    """

    # Manually passing in num_nodes, num_clusters, and num_graphs can boost the performance
    if num_nodes is None:
        if x is None:
            raise Exception("Please provide num_nodes if x is None")
        else:
            num_nodes = tf.shape(x)[0]

    assign_row, assign_col = assign_edge_index[0], assign_edge_index[1]

    transposed_sparse_assign_probs = tf.SparseTensor(
        indices=tf.cast(tf.stack([assign_col, assign_row], axis=1), dtype=tf.int64),
        values=assign_edge_weight,
        dense_shape=[num_clusters, num_nodes]
    )
    transposed_sparse_assign_probs = tf.sparse.reorder(transposed_sparse_assign_probs)

    transposed_sparse_adj = tf.SparseTensor(
        indices=tf.cast(tf.transpose(edge_index, [1, 0]), dtype=tf.int64),
        values=edge_weight,
        dense_shape=[num_nodes, num_nodes]
    )
    transposed_sparse_adj = tf.sparse.reorder(transposed_sparse_adj)

    # compute S'AS
    # Since tf only support sparse @ dense, we compute S'AS as S' @ (S' @ dense(A'))'
    pooled_adj = tf.sparse.sparse_dense_matmul(transposed_sparse_assign_probs, tf.sparse.to_dense(transposed_sparse_adj))
    pooled_adj = tf.sparse.sparse_dense_matmul(transposed_sparse_assign_probs, tf.transpose(pooled_adj, [1, 0]))

    pooled_edge_index, pooled_edge_weight = convert_dense_adj_to_edge(pooled_adj)

    # pooled_edge_index, pooled_edge_weight = remove_self_loop_edge(pooled_edge_index, pooled_edge_weight)

    if x is not None:
        pooled_x = tf.sparse.sparse_dense_matmul(transposed_sparse_assign_probs, x)
    else:
        pooled_x = None
    return pooled_x, pooled_edge_index, pooled_edge_weight



# coding=utf-8
from tf_geometric.datasets.ppi import PPIDataset
from tf_geometric.datasets.tu import TUDataset
from tf_geometric.datasets.planetoid import PlanetoidDataset, CoraDataset, CiteseerDataset, PubmedDataset, \
    SupervisedCoraDataset, SupervisedCiteseerDataset, SupervisedPubmedDataset


# coding=utf-8

# nn package contains functional APIs for tf_geometric
from tf_geometric.nn.kernel.map_reduce import *

from tf_geometric.nn.conv.gcn import gcn, gcn_norm_edge, gcn_cache_normed_edge
from tf_geometric.nn.conv.gat import gat
from tf_geometric.nn.conv.chebynet import chebynet, chebynet_norm_edge
from tf_geometric.nn.conv.sgc import sgc
from tf_geometric.nn.conv.tagcn import tagcn
from tf_geometric.nn.conv.graph_sage import mean_graph_sage, sum_graph_sage, mean_pool_graph_sage, max_pool_graph_sage, gcn_graph_sage, lstm_graph_sage
from tf_geometric.nn.conv.appnp import appnp
from tf_geometric.nn.conv.gin import gin
from tf_geometric.nn.conv.le_conv import le_conv


from tf_geometric.nn.sampling.drop_edge import drop_edge

from tf_geometric.nn.pool.common_pool import mean_pool, min_pool, max_pool, sum_pool
from tf_geometric.nn.pool.topk_pool import topk_pool
from tf_geometric.nn.pool.diff_pool import diff_pool, diff_pool_coarsen
from tf_geometric.nn.pool.set2set import set2set
from tf_geometric.nn.pool.cluster_pool import cluster_pool
from tf_geometric.nn.pool.sag_pool import sag_pool
from tf_geometric.nn.pool.asap import asap
from tf_geometric.nn.pool.sort_pool import sort_pool
from tf_geometric.nn.pool.min_cut_pool import min_cut_pool, min_cut_pool_coarsen, min_cut_pool_compute_losses



# coding=utf-8


import tensorflow as tf

from tf_geometric.nn.pool.min_cut_pool import min_cut_pool


class MinCutPool(tf.keras.Model):
    """
    OOP API for MinCutPool: "Spectral Clustering with Graph Neural Networks for Graph Pooling"
    """

    def __init__(self, feature_gnn, assign_gnn, units, num_clusters, activation=None, use_bias=True,
                 gnn_use_normed_edge=True,
                 bias_regularizer=None, *args, **kwargs):
        """
        MinCutPool

        :param feature_gnn: A GNN model to learn pooled node features, [x, edge_index, edge_weight] => updated_x,
            where updated_x corresponds to high-order node features.
        :param assign_gnn: A GNN model to learn cluster assignment for the pooling, [x, edge_index, edge_weight] => updated_x,
            where updated_x corresponds to the cluster assignment matrix.
        :param units: Positive integer, dimensionality of the output space. It must be provided if you set use_bias=True.
        :param num_clusters: Number of clusters for pooling.
        :param activation: Activation function to use.
        :param use_bias: Boolean, whether the layer uses a bias vector. If true, the "units" parameter must be provided.
        :param bias_regularizer: Regularizer function applied to the bias vector.
        """
        super().__init__(*args, **kwargs)
        self.feature_gnn = feature_gnn
        self.assign_gnn = assign_gnn

        self.num_clusters = num_clusters
        self.activation = activation

        if use_bias and units is None:
            raise Exception("The \"units\" parameter is required when you set use_bias=True.")

        if use_bias:
            self.bias = self.add_weight("bias", shape=[units],
                                        initializer="zeros", regularizer=bias_regularizer)

        self.gnn_use_normed_edge = gnn_use_normed_edge

    def call(self, inputs, cache=None, training=None, mask=None, return_losses=False):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight, node_graph_index]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Pooled graph: [pooled_x, pooled_edge_index, pooled_edge_weight, pooled_node_graph_index]
        """
        x, edge_index, edge_weight, node_graph_index = inputs

        outputs, loss_func = min_cut_pool(x, edge_index, edge_weight, node_graph_index,
                                          self.feature_gnn, self.assign_gnn, self.num_clusters,
                                          bias=self.bias, activation=self.activation,
                                          gnn_use_normed_edge=self.gnn_use_normed_edge,
                                          training=training, cache=cache,
                                          return_losses=True)
        self.add_loss(loss_func)

        if not return_losses:
            return outputs
        else:
            return outputs, loss_func



# coding=utf-8
from tf_geometric.nn.kernel.map_reduce import aggregate_neighbors, sum_reducer, identity_mapper, identity_updater


def gin_updater(x, reduced_neighbor_msg, eps):
    return x * (1.0 + eps) + reduced_neighbor_msg


def gin(x, edge_index, edge_weight, mlp_model, eps=0.0, training=None):
    """

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param mlp_model: A neural network (multi-layer perceptrons).
    :param eps: float, optional, (default: :obj:`0.`).
    :param training: Whether currently executing in training or inference mode.
    :return: Updated node features (x), shape: [num_nodes, num_output_features]
    """

    h = aggregate_neighbors(
        x, edge_index, edge_weight,
        identity_mapper,
        sum_reducer,
        identity_updater
    )

    h = gin_updater(x, h, eps)

    h = mlp_model(h, training=training)

    return h



# coding=utf-8
from tf_geometric.layers.conv.gcn import GCN
from tf_geometric.layers.conv.gat import GAT
from tf_geometric.layers.conv.gin import GIN
from tf_geometric.layers.conv.graph_sage import MeanGraphSage, SumGraphSage, MeanPoolGraphSage, MaxPoolGraphSage, GCNGraphSage, LSTMGraphSage
from tf_geometric.layers.conv.sgc import SGC
from tf_geometric.layers.conv.tagcn import TAGCN
from tf_geometric.layers.conv.chebynet import ChebyNet
from tf_geometric.layers.conv.appnp import APPNP
from tf_geometric.layers.conv.le_conv import LEConv

from tf_geometric.layers.sampling.drop_edge import DropEdge

from tf_geometric.layers.kernel.map_reduce import MapReduceGNN

from tf_geometric.layers.pool.common_pool import MeanPool, MinPool, MaxPool, SumPool
from tf_geometric.layers.pool.diff_pool import DiffPool
from tf_geometric.layers.pool.set2set import Set2Set
from tf_geometric.layers.pool.sag_pool import SAGPool
from tf_geometric.layers.pool.asap import ASAP
from tf_geometric.layers.pool.sort_pool import SortPool
from tf_geometric.layers.pool.min_cut_pool import MinCutPool



# coding=utf-8
import tensorflow as tf
import pickle



def download_file(path, url_or_urls):
    if not isinstance(url_or_urls, list):
        urls = [url_or_urls]
    else:
        urls = url_or_urls

    last_except = None
    for url in urls:
        try:
            return tf.keras.utils.get_file(path, origin=url)
        except Exception as e:
            last_except = e
            print(e)

    raise last_except


def save_cache(obj, path):
    with open(path, "wb") as f:
        pickle.dump(obj, f)


def load_cache(path):
    # if not os.path.exists(path):
    #     return None

    with open(path, "rb") as f:
        return pickle.load(f)


# coding=utf-8
import tensorflow as tf
import numpy as np
import networkx as nx
from sklearn.model_selection import train_test_split
import warnings

from tf_geometric.utils.union_utils import convert_union_to_numpy

from scipy.sparse.linalg import eigs, eigsh
import scipy.sparse


def convert_x_to_3d(x, source_index, K=None, pad=True):

    source_index_perm = tf.argsort(source_index, stable=True)
    sorted_source_index = tf.gather(source_index, source_index_perm)
    permed_x = tf.gather(x, source_index_perm)

    num_targets = tf.shape(sorted_source_index)[0]
    target_ones = tf.ones([num_targets], dtype=tf.int32)
    num_targets_for_sources = tf.math.segment_sum(target_ones, sorted_source_index)
    max_num_targets_for_sources = tf.reduce_max(num_targets_for_sources)

    # max index of source + 1
    num_seen_sources = tf.shape(num_targets_for_sources)[0]

    num_targets_before = tf.concat([
        tf.zeros([1], dtype=tf.int32),
        tf.math.cumsum(num_targets_for_sources)[:-1]
    ], axis=0)

    target_index_for_source = tf.range(0, num_targets) - tf.gather(num_targets_before, sorted_source_index)

    if K is None:
        K = max_num_targets_for_sources
    elif K > max_num_targets_for_sources:
        if not pad:
            K = max_num_targets_for_sources
    elif K < max_num_targets_for_sources:
            mask = tf.less(target_index_for_source, K)
            target_index_for_source = tf.boolean_mask(target_index_for_source, mask)
            sorted_source_index = tf.boolean_mask(sorted_source_index, mask)
            permed_x = tf.boolean_mask(permed_x, mask)


    h = tf.zeros([num_seen_sources, K, tf.shape(x)[-1]], dtype=x.dtype)
    index = tf.stack([sorted_source_index, target_index_for_source], axis=1)
    h = tf.tensor_scatter_nd_update(h, index, permed_x)
    return h


def remove_self_loop_edge(edge_index, edge_weight=None):
    edge_index_is_tensor = tf.is_tensor(edge_index)
    edge_weight_is_tensor = tf.is_tensor(edge_weight)

    row, col = edge_index[0], edge_index[1]
    mask = tf.not_equal(row, col)

    edge_index = tf.boolean_mask(edge_index, mask, axis=1)
    if edge_weight is not None:
        edge_weight = tf.boolean_mask(edge_weight, mask, axis=0)

    if not edge_index_is_tensor:
        edge_index = edge_index.numpy()

    if edge_weight is not None and not edge_weight_is_tensor:
        edge_weight = edge_weight.numpy()

    return edge_index, edge_weight


def convert_dense_adj_to_edge(dense_adj):
    num_nodes = tf.shape(dense_adj)[0]
    row, col = tf.meshgrid(tf.range(num_nodes), tf.range(num_nodes), indexing="ij")
    row = tf.reshape(row, [-1])
    col = tf.reshape(col, [-1])
    edge_index = tf.stack([row, col], axis=0)
    edge_weight = tf.reshape(dense_adj, [-1])

    edge_mask = tf.not_equal(edge_weight, 0.0)
    edge_index = tf.boolean_mask(edge_index, edge_mask, axis=1)
    edge_weight = tf.boolean_mask(edge_weight, edge_mask)

    return edge_index, edge_weight


def convert_dense_assign_to_edge(dense_assign, node_graph_index=None, num_nodes=None, num_clusters=None):
    """
    Convert a dense assignment matrix [num_nodes, num_clusters] of a Graph or BatchGraph to edge_index, edge_weight.
    In the single-graph scenario (when node_graph_index is None), there are num_clusters clusters and 
        the j_th column of each row correspond to the j_th clusters.
    In the multi-graph scenario (where node_graph_index is not None), there are num_clusters * num_graphs clusters and
        the j_th column of the i_th row correspond to the j_th cluster in the corresponding graph, and the cluster index
        is num_clusters * node_graph_index[i] + j
    
    :param dense_assign: 
    :param node_graph_index: 
    :param num_nodes: 
    :param num_clusters: 
    :return: 
    """
    dense_shape = tf.shape(dense_assign)

    # manually passing in num_row and num_col can boost the computation
    if num_nodes is None:
        num_nodes = dense_shape[0]
    if num_clusters is None:
        num_clusters = dense_shape[1]

    row, col = tf.meshgrid(tf.range(num_nodes), tf.range(num_clusters), indexing="ij")

    # multi graphs, the j_th clusters of the i_th graph is assigned a new cluster index: num_clusters * i + j
    if node_graph_index is not None:
        col += tf.expand_dims(node_graph_index, axis=-1) * num_clusters

    row = tf.reshape(row, [-1])
    col = tf.reshape(col, [-1])
    edge_index = tf.stack([row, col], axis=0)

    edge_weight = tf.reshape(dense_assign, [-1])

    return edge_index, edge_weight


def convert_edge_to_nx_graph(edge_index, edge_properties=[], convert_to_directed=False):
    edge_index = convert_union_to_numpy(edge_index, dtype=np.int32)
    edge_properties = [convert_union_to_numpy(edge_property) for edge_property in edge_properties]

    g = nx.Graph()
    for i in range(edge_index.shape[1]):
        property_dict = {
        }

        for j, edge_property in enumerate(edge_properties):
            if edge_property is not None:
                property_dict["p_{}".format(j)] = edge_property[i]

        g.add_edge(edge_index[0, i], edge_index[1, i], **property_dict)

    if convert_to_directed:
        g = g.to_directed()

    return g


def convert_edge_to_upper(edge_index, edge_properties=[]):
    edge_index_is_tensor = tf.is_tensor(edge_index)
    edge_properties_is_tensor = [tf.is_tensor(edge_property) for edge_property in edge_properties]

    g = convert_edge_to_nx_graph(edge_index, edge_properties, convert_to_directed=False)

    sorted_edges = [sorted(edge) for edge in g.edges]
    edge_index = np.array(sorted_edges).T

    edge_properties = [
        np.array([item[2] for item in g.edges.data("p_{}".format(i))])
        if edge_property is not None else None
        for i, edge_property in enumerate(edge_properties)
    ]

    if edge_index_is_tensor:
        edge_index = tf.convert_to_tensor(edge_index)

    edge_properties = [
        tf.convert_to_tensor(edge_property) if edge_properties_is_tensor[i] else edge_property
        for i, edge_property in enumerate(edge_properties)
    ]

    return edge_index, edge_properties


# [[1,3,5], [2,1,4]] => [[1,3,5,2,1,4], [2,1,4,1,3,5]]
def convert_edge_to_directed(edge_index, edge_properties=[]):
    edge_index_is_tensor = tf.is_tensor(edge_index)
    edge_properties_is_tensor = [tf.is_tensor(edge_property) for edge_property in edge_properties]

    g = convert_edge_to_nx_graph(edge_index, edge_properties, convert_to_directed=True)

    edge_index = np.array(g.edges).T

    edge_properties = [
        np.array([item[2] for item in g.edges.data("p_{}".format(i))])
        if edge_property is not None else None
        for i, edge_property in enumerate(edge_properties)
    ]

    if edge_index_is_tensor:
        edge_index = tf.convert_to_tensor(edge_index)

    edge_properties = [
        tf.convert_to_tensor(edge_property) if edge_properties_is_tensor[i] else edge_property
        for i, edge_property in enumerate(edge_properties)
    ]

    return edge_index, edge_properties


def add_self_loop_edge(edge_index, num_nodes, edge_weight=None, fill_weight=1.0):
    diagnal_edge_index = tf.stack([tf.range(num_nodes, dtype=tf.int32)] * 2, axis=0)
    updated_edge_index = tf.concat([edge_index, diagnal_edge_index], axis=1)

    if not tf.is_tensor(edge_index):
        updated_edge_index = updated_edge_index.numpy()

    if edge_weight is not None:
        diagnal_edge_weight = tf.cast(tf.fill([num_nodes], fill_weight), tf.float32)
        updated_edge_weight = tf.concat([edge_weight, diagnal_edge_weight], axis=0)

        if not tf.is_tensor(edge_weight):
            updated_edge_weight = updated_edge_weight.numpy()
    else:
        updated_edge_weight = None

    return updated_edge_index, updated_edge_weight


def negative_sampling(num_samples, num_nodes, edge_index=None, replace=True, mode="undirected",
                      batch_size=None):
    """

    :param num_samples:
    :param num_nodes:
    :param edge_index: if edge_index is provided, sampled positive edges will be filtered
    :param replace: only works when edge_index is provided, deciding whether sampled edges should be unique
    :param if batch_size is None, return edge_index, otherwise return a list of batch_size edge_index
    :return:
    """

    edge_index = convert_union_to_numpy(edge_index, np.int32)
    fake_batch_size = 1 if batch_size is None else batch_size

    if edge_index is None:
        sampled_edge_index_list = [np.random.randint(0, num_nodes, [2, num_samples]).astype(np.int32)
                                   for _ in range(fake_batch_size)]
    else:
        if mode == "undirected":
            # fast
            edge_index, _ = convert_edge_to_upper(edge_index)
            adj = np.ones([num_nodes, num_nodes])
            # np.fill_diagonal(adj, 0)
            adj = np.triu(adj, k=1)
            adj[edge_index[0], edge_index[1]] = 0
            neg_edges = np.nonzero(adj)
            neg_edge_index = np.stack(neg_edges, axis=0)
            sampled_edge_index_list = []
            for _ in range(fake_batch_size):
                random_indices = np.random.choice(list(range(neg_edge_index.shape[1])), num_samples, replace=replace)
                sampled_edge_index = neg_edge_index[:, random_indices].astype(np.int32)
                sampled_edge_index_list.append(sampled_edge_index)
        else:
            raise NotImplementedError()

    if tf.is_tensor(edge_index):
        sampled_edge_index_list = [tf.convert_to_tensor(sampled_edge_index)
                                   for sampled_edge_index in sampled_edge_index_list]

    if batch_size is None:
        return sampled_edge_index_list[0]
    else:
        return sampled_edge_index_list


def negative_sampling_with_start_node(start_node_index, num_nodes, edge_index=None):
    """

    :param start_node_index: Tensor or ndarray
    :param num_nodes:
    :param edge_index: if edge_index is provided, sampled positive edges will be filtered
    :return:
    """

    start_node_index_is_tensor = tf.is_tensor(start_node_index)

    start_node_index = convert_union_to_numpy(start_node_index, dtype=np.int32)
    edge_index = convert_union_to_numpy(edge_index, np.int32)
    num_samples = len(start_node_index)

    if edge_index is None:
        end_node_index = np.random.randint(0, num_nodes, [num_samples]).astype(np.int32)
        sampled_edge_index = np.stack([start_node_index, end_node_index], axis=0)
    else:
        edge_set = set([tuple(edge) for edge in edge_index.T])

        sampled_edges = []
        for a in start_node_index:
            while True:
                b = np.random.randint(0, num_nodes, dtype=np.int32)
                if a == b:
                    continue
                edge = (a, b)
                if edge not in edge_set:
                    sampled_edges.append(edge)
                    break

        sampled_edge_index = np.array(sampled_edges, dtype=np.int32).T

    if start_node_index_is_tensor:
        sampled_edge_index = tf.convert_to_tensor(sampled_edge_index)

    return sampled_edge_index


def extract_unique_edge(edge_index, edge_weight=None, mode="undirected"):
    is_edge_index_tensor = tf.is_tensor(edge_index)
    is_edge_weight_tensor = tf.is_tensor(edge_weight)

    edge_index = convert_union_to_numpy(edge_index, dtype=np.int32)
    edge_weight = convert_union_to_numpy(edge_weight, dtype=np.float32)

    edge_set = set()
    unique_edge_index = []
    for i in range(edge_index.shape[1]):
        edge = edge_index[:, i]
        if mode == "undirected":
            edge = sorted(edge)
        edge = tuple(edge)

        if edge in edge_set:
            continue
        else:
            unique_edge_index.append(i)
            edge_set.add(edge)

    edge_index = edge_index[:, unique_edge_index]
    if is_edge_index_tensor:
        edge_index = tf.convert_to_tensor(edge_index)

    if edge_weight is not None:
        edge_weight = edge_weight[unique_edge_index]
        if is_edge_weight_tensor:
            edge_weight = tf.convert_to_tensor(edge_weight)

    return edge_index, edge_weight


def edge_train_test_split(edge_index, test_size, edge_weight=None, mode="undirected", **kwargs):
    """

    :param edge_index:
    :param test_size:
    :param edge_weight:
    :param mode:
    :return:
    """

    # todo: warn user if they pass into "num_nodes", deprecated
    if "num_nodes" in kwargs:
        warnings.warn(
            "argument \"num_nodes\" is deprecated for the method \"edge_train_test_split\", you can remove it")

    if mode == "undirected":
        is_edge_index_tensor = tf.is_tensor(edge_index)
        is_edge_weight_tensor = tf.is_tensor(edge_weight)

        edge_index = convert_union_to_numpy(edge_index, dtype=np.int32)
        edge_weight = convert_union_to_numpy(edge_weight, dtype=np.float32)

        upper_edge_index, [upper_edge_weight] = convert_edge_to_upper(edge_index, [edge_weight])

        num_unique_edges = upper_edge_index.shape[1]
        train_indices, test_indices = train_test_split(list(range(num_unique_edges)), test_size=test_size)
        undirected_train_edge_index = upper_edge_index[:, train_indices]
        undirected_test_edge_index = upper_edge_index[:, test_indices]

        if is_edge_index_tensor:
            undirected_train_edge_index = tf.convert_to_tensor(undirected_train_edge_index)
            undirected_test_edge_index = tf.convert_to_tensor(undirected_test_edge_index)

        if edge_weight is not None:
            undirected_train_edge_weight = upper_edge_weight[train_indices]
            undirected_test_edge_weight = upper_edge_weight[test_indices]

            if is_edge_weight_tensor:
                undirected_train_edge_weight = tf.convert_to_tensor(undirected_train_edge_weight)
                undirected_test_edge_weight = tf.convert_to_tensor(undirected_test_edge_weight)
        else:
            undirected_train_edge_weight = None
            undirected_test_edge_weight = None

        return undirected_train_edge_index, undirected_test_edge_index, undirected_train_edge_weight, undirected_test_edge_weight

    else:
        raise NotImplementedError()


def compute_edge_mask_by_node_index(edge_index, node_index):
    edge_index_is_tensor = tf.is_tensor(edge_index)

    max_node_index = tf.maximum(tf.reduce_max(edge_index), tf.reduce_max(node_index))
    node_mask = tf.scatter_nd(tf.expand_dims(node_index, axis=-1), tf.ones_like(node_index), [max_node_index + 1])
    node_mask = tf.cast(node_mask, tf.bool)
    row, col = edge_index[0], edge_index[1]
    row_mask = tf.gather(node_mask, row)
    col_mask = tf.gather(node_mask, col)
    edge_mask = tf.logical_and(row_mask, col_mask)

    if not edge_index_is_tensor:
        edge_mask = edge_mask.numpy()
    return edge_mask


def get_laplacian(edge_index, num_nodes, edge_weight, normalization_type, fill_weight=1.0):
    if normalization_type is not None:
        assert normalization_type in [None, 'sym', 'rw']

    row, col = edge_index[0], edge_index[1]
    deg = tf.math.unsorted_segment_sum(edge_weight, row, num_segments=num_nodes)
    ##L = D - A
    if normalization_type is None:
        edge_index, edge_weight = add_self_loop_edge(edge_index, num_nodes, edge_weight, fill_weight=fill_weight)
        row, col = edge_index
        deg_inv = tf.where(
            tf.math.logical_or(tf.math.is_inf(deg), tf.math.is_nan(deg)),
            tf.zeros_like(deg),
            deg
        )
        edge_weight = tf.gather(deg_inv, row) - edge_weight

    ## L^ = D^{-1/2}LD^{-1/2}
    elif normalization_type == 'sym':
        deg_inv_sqrt = tf.pow(deg, -0.5)
        deg_inv_sqrt = tf.where(
            tf.math.logical_or(tf.math.is_inf(deg_inv_sqrt), tf.math.is_nan(deg_inv_sqrt)),
            tf.zeros_like(deg_inv_sqrt),
            deg_inv_sqrt
        )

        normed_edge_weight = tf.gather(deg_inv_sqrt, row) * edge_weight * tf.gather(deg_inv_sqrt, col)
        edge_index, tmp = add_self_loop_edge(edge_index, num_nodes, edge_weight=normed_edge_weight,
                                             fill_weight=fill_weight)

        assert tmp is not None
        edge_weight = tmp
    ##L^ = D^{-1}L
    else:
        deg_inv = 1.0 / deg
        deg_inv = tf.where(
            tf.math.logical_or(tf.math.is_inf(deg_inv), tf.math.is_nan(deg_inv)),
            tf.zeros_like(deg_inv),
            deg_inv
        )

        normed_edge_weight = tf.gather(deg_inv, row) * edge_weight

        edge_index, tmp = add_self_loop_edge(edge_index, num_nodes, edge_weight=normed_edge_weight,
                                             fill_weight=fill_weight)

        assert tmp is not None
        edge_weight = tmp

    return edge_index, edge_weight


def to_scipy_sparse_matrix(edge_index, edge_weight=None, num_nodes=None):
    r"""Converts a graph given by edge indices and edge attributes to a scipy
    sparse matrix.

    Args:
        edge_index (LongTensor): The edge indices.
        edge_attr (Tensor, optional): Edge weights or multi-dimensional
            edge features. (default: :obj:`None`)
        num_nodes (int, optional): The number of nodes, *i.e.*
            :obj:`max_val + 1` of :attr:`index`. (default: :obj:`None`)
    """
    row, col = edge_index

    if edge_weight is None:
        edge_weight = tf.ones(row.shape[0])
    else:
        edge_weight = tf.reshape(edge_weight, [-1])
        assert edge_weight.shape[0] == row.shape[0]

    N = num_nodes
    out = scipy.sparse.coo_matrix((edge_weight, (row, col)), (N, N))
    return out


class RandomNeighborSampler(object):
    def __init__(self, edge_index, edge_weight=None):
        self.edge_index = convert_union_to_numpy(edge_index, np.int32)
        if edge_weight is not None:
            self.edge_weight = convert_union_to_numpy(edge_weight)
        else:
            self.edge_weight = np.ones([self.edge_index.shape[1]], dtype=np.float32)
        self.neighbor_dict = {}

        for (a, b), weight in zip(self.edge_index.T, self.edge_weight):

            if a not in self.neighbor_dict:
                neighbors = []
                self.neighbor_dict[a] = neighbors
            else:
                neighbors = self.neighbor_dict[a]
            neighbors.append((b, weight))

        self.num_sources = len(self.neighbor_dict)
        self.source_index = sorted(self.neighbor_dict.keys())
        self.neighbors_list = [self.neighbor_dict[a] for a in self.source_index]
        self.num_neighbors_list = np.array([len(neighbors) for neighbors in self.neighbors_list])
        self.neighbor_index_list = [np.arange(num_neighbors) for num_neighbors in self.num_neighbors_list]

    def sample(self, k=None, ratio=None):
        if k is None and ratio is None:
            raise Exception("you should provide either k or ratio")
        elif k is not None and ratio is not None:
            raise Exception("you should provide either k or ratio, not both of them")

        if ratio is not None:
            num_sampled_neighbors = np.ceil(self.num_neighbors_list * ratio).astype(np.int32)
        else:
            num_sampled_neighbors = np.full([self.num_sources], fill_value=k)

        sampled_edge_index = []
        sampled_edge_weight = []

        for i, (a, neighbors, num_sampled_neighbors, neighbor_index) in \
                enumerate(zip(self.source_index, self.neighbors_list, num_sampled_neighbors, self.neighbor_index_list)):
            # sampled_neighbors = np.random.choice(neighbors, num_sampled_neighbors, replace=True)
            sampled_neighbor_index = np.random.choice(neighbor_index, num_sampled_neighbors, replace=True)
            sampled_neighbors = [neighbors[i] for i in sampled_neighbor_index]

            for (b, weight) in sampled_neighbors:
                sampled_edge_index.append([a, b])
                sampled_edge_weight.append(weight)

        sampled_edge_index = np.array(sampled_edge_index).T
        sampled_edge_weight = np.array(sampled_edge_weight)
        return sampled_edge_index, sampled_edge_weight


class LaplacianMaxEigenvalue(object):
    def __init__(self, edge_index, num_nodes, edge_weight, is_undirected=True):
        self.num_nodes = num_nodes
        self.edge_index = convert_union_to_numpy(edge_index, np.int32)
        if edge_weight is not None:
            self.edge_weight = convert_union_to_numpy(edge_weight)
        else:
            self.edge_weight = np.ones([self.edge_index.shape[1]], dtype=np.float32)
        self.is_undirected = is_undirected

    def __call__(self, normalization_type='sym'):
        assert normalization_type in [None, 'sym', 'rw']

        edge_index, edge_weight = remove_self_loop_edge(self.edge_index, self.edge_weight)

        edge_index, edge_weight = get_laplacian(self.edge_index, self.num_nodes, edge_weight, normalization_type)

        L = to_scipy_sparse_matrix(edge_index, edge_weight, self.num_nodes)

        eig_fn = eigs
        if self.is_undirected and normalization_type:
            eig_fn = eigsh

        lambda_max = eig_fn(L, k=1, which='LM', return_eigenvectors=False)

        return float(lambda_max.real)


# CACHE_KEY_ADJ_NORMED_EDGE = "adj_normed_edge"

def adj_norm_edge(edge_index, num_nodes, edge_weight=None, add_self_loop=False, cache=None):
    cache_key = "adj_normed_edge"
    if cache is not None:
        cached_data = cache.get(cache_key, None)
        if cached_data is not None:
            return cached_data

    if edge_weight is None:
        edge_weight = tf.ones([tf.shape(edge_index)[1]], dtype=tf.float32)

    if add_self_loop:
        fill_weight = 1.0
        edge_index, edge_weight = add_self_loop_edge(edge_index, num_nodes, edge_weight=edge_weight,
                                                     fill_weight=fill_weight)

    row, col = edge_index[0], edge_index[1]
    deg = tf.math.unsorted_segment_sum(edge_weight, row, num_segments=num_nodes)
    deg_inv_sqrt = tf.pow(deg, -0.5)
    deg_inv_sqrt = tf.where(
        tf.math.logical_or(tf.math.is_inf(deg_inv_sqrt), tf.math.is_nan(deg_inv_sqrt)),
        tf.zeros_like(deg_inv_sqrt),
        deg_inv_sqrt
    )

    normed_edge_weight = tf.gather(deg_inv_sqrt, row) * edge_weight * tf.gather(deg_inv_sqrt, col)

    if cache is not None:
        cache[cache_key] = edge_index, normed_edge_weight

    return edge_index, normed_edge_weight


# coding=utf-8
import tensorflow as tf
from tf_geometric.utils.union_utils import union_len


def topk_pool(source_index, score, K=None, ratio=None):
    """

    :param source_index: index of source node (of edge) or source graph (of node)
    :param score: 1-D Array
    :param K: Keep top K targets for each source
    :param ratio: Keep num_targets * ratio targets for each source
    :return: sampled_edge_index, sampled_edge_score, sample_index
    """

    if K is None and ratio is None:
        raise Exception("you should provide either k or ratio for topk_pool")
    elif K is not None and ratio is not None:
        raise Exception("you should provide either k or ratio for topk_pool, not both of them")

    # currently, we consider the source_index is not sorted
    # the option is preserved for future performance optimization
    source_index_sorted = False

    if source_index_sorted:
        sorted_source_index = source_index
        # sort score by source_index
        sorted_score = score
    else:
        source_index_perm = tf.argsort(source_index)
        sorted_source_index = tf.gather(source_index, source_index_perm)
        sorted_score = tf.gather(score, source_index_perm)

    sorted_score = tf.reshape(sorted_score, [-1])

    num_targets = tf.shape(sorted_source_index)[0]
    target_ones = tf.ones([num_targets], dtype=tf.int32)
    num_targets_for_sources = tf.math.segment_sum(target_ones, sorted_source_index)
    # number of columns for score matrix
    num_cols = tf.reduce_max(num_targets_for_sources)

    # max index of source + 1
    num_seen_sources = tf.shape(num_targets_for_sources)[0]

    min_score = tf.reduce_min(sorted_score)

    num_targets_before = tf.concat([
        tf.zeros([1], dtype=tf.int32),
        tf.math.cumsum(num_targets_for_sources)[:-1]
    ], axis=0)

    target_index_for_source = tf.range(0, num_targets) - tf.gather(num_targets_before, sorted_source_index)

    score_matrix = tf.cast(tf.fill([num_seen_sources, num_cols], min_score - 1.0), dtype=tf.float32)
    score_index = tf.stack([sorted_source_index, target_index_for_source], axis=1)
    score_matrix = tf.tensor_scatter_nd_update(score_matrix, score_index, sorted_score)

    sort_index = tf.argsort(score_matrix, axis=-1, direction="DESCENDING")

    if K is not None:
        node_k = tf.math.minimum(
            tf.cast(tf.fill([num_seen_sources], K), dtype=tf.int32),
            num_targets_for_sources
        )
    else:
        node_k = tf.cast(
            tf.math.ceil(tf.cast(num_targets_for_sources, dtype=tf.float32) * tf.cast(ratio, dtype=tf.float32)),
            dtype=tf.int32
        )

    row, col = tf.meshgrid(tf.range(num_seen_sources), tf.range(tf.reduce_max(node_k)), indexing="ij")
    row = tf.reshape(row, [-1])
    col = tf.reshape(col, [-1])
    repeated_k = tf.gather(node_k, row)
    k_mask = tf.less(col, repeated_k)

    row = tf.boolean_mask(row, k_mask)
    col = tf.boolean_mask(col, k_mask)

    sample_col_index = tf.gather_nd(sort_index, tf.stack([row, col], axis=1))

    topk_index = tf.gather(num_targets_before, row) + sample_col_index

    if source_index_sorted:
        return topk_index
    else:
        return tf.gather(source_index_perm, topk_index)



from setuptools import setup, find_packages

setup(
    name="tf_geometric",
    python_requires='>3.5.0',
    version="0.0.47",
    author="Jun Hu",
    author_email="hujunxianligong@gmail.com",
    packages=find_packages(
        exclude=[
            'data',
            'demo',
            'dist',
            'doc',
            'docs',
            'logs',
            'models',
            'test'
        ]
    ),
    install_requires=[
        "numpy >= 1.17.4",
        "networkx >= 2.1",
        "scipy >= 1.1.0",
        "scikit-learn >= 0.22",
        "tqdm"
    ],
    extras_require={
        'tf1-cpu': ["tensorflow >= 1.14.0,<2.0.0"],
        'tf1-gpu': ["tensorflow-gpu >= 1.14.0,<2.0.0"],
        'tf2-cpu': ["tensorflow >= 2.0.0b1"],
        'tf2-gpu': ["tensorflow-gpu >= 2.0.0b1"]
    },
    description="""
        Efficient and Friendly Graph Neural Network Library for TensorFlow 1.x and 2.x.
    """,
    license="GNU General Public License v3.0 (See LICENSE)",
    long_description=open("README.rst", "r", encoding="utf-8").read(),
    url="https://github.com/CrawlScript/tf_geometric"
)